/*
 * Copyright (C) 2004-2012 Charles Cazabon <charlesc-memtester@pyropus.ca>
 * Licensed under the terms of the GNU General Public License version 2 (only).
 * Copyright 2019-2022 RnD Center "ELVEES", JSC
 */

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#ifndef BUILD_FOR_ELCORECL
#include <error.h>
#include <sys/random.h>
#else
/* TODO: Replace with <sys/random.h> after adding this header
 * file to the toolchain.
 */
ssize_t getrandom(void *buf, size_t buflen, unsigned int flags);
#include "dsp-funcs.h"
#endif

#include "types.h"
#include "sizes.h"

#define min(a,b) \
   ({ __typeof__ (a) _a = (a); \
      __typeof__ (b) _b = (b); \
      _a > _b ? _b : _a; })

extern int id;
extern int verbose;
extern int stop_on_error;
extern ulv available_bit_mask;

int compare_regions_bfailure(ulv * bufa, ulv * bufb, size_t count)
{
    int r = 0;
    size_t i;
    ulv *p1 = bufa;
    ulv *p2 = bufb;

    for (i = 0; i < count; i++, p1++, p2++) {
        if ((*p1 & available_bit_mask) != (*p2 & available_bit_mask)) {
            r += 1;
            if (verbose) {
                error(0, 0,
                      "memtester-simple #%d: FAILURE: 0x%08lx != 0x%08lx at offset 0x%08lx.\n",
                      id, (ul) *p1 & available_bit_mask, (ul) *p2 & available_bit_mask, (ul) (i * sizeof(ul)));

                if ((((*p1 & available_bit_mask) >> 24) & 0xff) !=
                                (((*p2 & available_bit_mask) >> 24) & 0xff))
                    error(0, 0, "#%d: failure in byte 3\n", id);
                if ((((*p1 & available_bit_mask) >> 16) & 0xff) !=
                                (((*p2 & available_bit_mask) >> 16) & 0xff))
                    error(0, 0, "#%d: failure in byte 2\n", id);
                if ((((*p1 & available_bit_mask) >> 8) & 0xff) !=
                                (((*p2 & available_bit_mask) >> 8) & 0xff))
                    error(0, 0, "#%d: failure in byte 1\n", id);
                if ((((*p1 & available_bit_mask) >> 0) & 0xff) !=
                                (((*p2 & available_bit_mask) >> 0) & 0xff))
                    error(0, 0, "#%d: failure in byte 0\n", id);
            }

            if (stop_on_error)
                return r;
        }
    }
    return r;
}

int test_solidbits_comparison(ulv * bufa, ulv * bufb, size_t count, int cycles,
                              int cmp_num)
{
    ulv *p1 = bufa;
    ulv *p2 = bufb;
    unsigned int j, k, errors, was_err = 0, total_errors = 0;
    ul q;
    size_t i;
    for (j = 0; j < cycles; j++) {
        q = (j % 2) == 0 ? UL_ONEBITS : 0;
        p1 = (ulv *) bufa;
        p2 = (ulv *) bufb;
        for (i = 0; i < count; i++) {
            *p1++ = *p2++ = (i % 2) == 0 ? q : ~q;
        }
        for (k = 0; k < cmp_num; k++) {
            errors = compare_regions_bfailure(bufa, bufb, count);
            if (stop_on_error && errors)
                return -1;
            if (errors) {
                total_errors += errors;
                was_err = 1;
            }
        }
    }
    printf("memtester-simple #%d: Total errors %d\n", id, total_errors);
    if (was_err)
        return -1;
    else
        return 0;
}

int test_bitflip_comparison(ulv * bufa, ulv * bufb, size_t count, int cycles,
                            int cmp_num)
{
    ulv *p1 = bufa;
    ulv *p2 = bufb;
    unsigned int j, k, errors, was_err = 0, total_errors = 0;
    ul q;
    size_t i;

    for (k = 0; k < UL_LEN; k++) {
        q = ONE << k;
        for (j = 0; j < cycles; j++) {
            q = ~q;
            p1 = (ulv *) bufa;
            p2 = (ulv *) bufb;
            for (i = 0; i < count; i++) {
                *p1++ = *p2++ = (i % 2) == 0 ? q : ~q;
            }

            for (i = 0; i < cmp_num; i++) {
                errors = compare_regions_bfailure(bufa, bufb, count);
                if (stop_on_error && errors)
                    return -1;
                if (errors) {
                    total_errors += errors;
                    was_err = 1;
                }
            }

        }
    }

    printf("memtester-simple #%d: Total errors %d\n", id, total_errors);
    if (was_err)
        return -1;
    else
        return 0;
}

int test_randdata(ulv * bufa, ulv * bufb, size_t count, int cycles, int cmp_num)
{
    ulv *p1 = bufa;
    ulv *p2 = bufb;
    unsigned int j, k, errors, was_err = 0, total_errors = 0;
    size_t i;
    ul nbytes;
    ssize_t rand_bytes_requested, rand_bytes_got;

    for (j = 0; j < cycles; j++) {
        p1 = (ulv *) bufa;
        p2 = (ulv *) bufb;
        nbytes = count * sizeof(ul);

        while (nbytes > 0) {
            /* getrandom() returns up to (32MB - 1byte) random data */
            rand_bytes_requested = min(31 * 1024 * 1024, nbytes);
            /* TODO: Replace with rand() that takes SEED from env variable */
            rand_bytes_got = getrandom((void *)p1, rand_bytes_requested, 0);
            if (rand_bytes_got != rand_bytes_requested)
                return -1;

            nbytes -= rand_bytes_got;
            for (i = 0; i < rand_bytes_got / sizeof(ul); i++)
                *p2++ = *p1++;
        }

        for (k = 0; k < cmp_num; k++) {
            errors = compare_regions_bfailure(bufa, bufb, count);

            if (stop_on_error && errors)
                return -1;

            if (errors) {
                total_errors += errors;
                was_err = 1;
            }
        }
    }

    printf("memtester-simple #%d: Total errors %d\n", id, total_errors);

    if (was_err)
        return -1;
    else
        return 0;
}
